# -*- coding: utf-8 -*-
#
# Copyright © 2012 CEA
# Pierre Raybaut
# Licensed under the terms of the CECILL License
# (see guidata/__init__.py for details)

"""
guiqwt.widgets
==============

The ``guiqwt.widgets`` package provides ready-to-use widgets for creating 
signal/image processing GUI applications.

.. automodule:: guiqwt.widgets.fit
   :members:
   :inherited-members:

.. automodule:: guiqwt.widgets.resizedialog
   :members:
   :inherited-members:
       
.. automodule:: guiqwt.widgets.rotatecrop
   :members:
   :inherited-members:
"""
